
/******************************************************************************
* MODULE     : notify.gen.cc
* DESCRIPTION: undertake the appropriate action after
*              notification of changes in edit tree
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <data.gen.h>
#include <edit_modify.gen.h>
#include <edit_cursor.gen.h>
#include <tm_buffer.gen.h>

#module code_notify
#import data
#import edit_modify
#import edit_cursor
#import tm_buffer

extern int max_undo_depth;

/******************************************************************************
* Constructors and destructors
******************************************************************************/

edit_modify::edit_modify (tm_buffer buf):
  edit_base (buf), undo_flag (FALSE), redo_flag (FALSE) {}
edit_modify::~edit_modify () {}

/******************************************************************************
* modification routines
******************************************************************************/

void
edit_modify::assign (path pp, tree u) {
  path p= copy (pp);
  // cout << "Assign " << u << " at " << p << "\n";
  notify_undo ("assign", p, subtree (et, p));

  int i;
  for (i=0; i<N(buf->vws); i++)
    ((tm_view) (buf->vws[i]))->ed->notify_assign (et, p, u);

  subtree (et, p)= u;
  finished ();
}

void
edit_modify::insert (path pp, tree u) {
  path p= copy (pp);
  // cout << "Insert " << u << " at " << p << "\n";
  notify_undo ("remove", p, as_string (N(u)==0? N(u->label): N(u)));

  int i;
  for (i=0; i<N(buf->vws); i++)
    ((tm_view) (buf->vws[i]))->ed->notify_insert (et, p, u);

  insert_at (et, p, u);
  finished ();
}

void
edit_modify::remove (path pp, int nr) {
  path p= copy (pp);
  // cout << "Remove " << nr << " at " << p << "\n";
  tree& st (subtree (et, path_up (p)));
  int l= last_item (p);
  if (N(st)==0) notify_undo ("insert", p, st->label (l, l+ nr));
  else notify_undo ("insert", p, st (l, l+ nr));

  int i;
  for (i=0; i<N(buf->vws); i++)
    ((tm_view) (buf->vws[i]))->ed->notify_remove (et, p, nr);

  remove_at (et, p, nr);
  finished ();
}

void
edit_modify::split (path pp) {
  path p= copy (pp);
  // cout << "Split at " << p << "\n";
  if (N(p)<2) fatal_error ("path too short in split", "editor::split");
  tree& st (subtree (et, path_up (path_up (p))));
  int  l1 = last_item (path_up (p));
  int  l2 = last_item (p);
  notify_undo ("join", path_up (p), "");

  int i;
  for (i=0; i<N(buf->vws); i++)
    ((tm_view) (buf->vws[i]))->ed->notify_split (et, p);

  if (N(st[l1])==0) {
    string s1, s2;
    ::split (st[l1]->label, l2, s1, s2);
    st[l1]= s2;
    st= insert_one (st, l1, tree (s1));
  }
  else {
    tree st1, st2;
    ::split (st[l1], l2, st1, st2);
    st[l1]= st2;
    st= insert_one (st, l1, st1);
  }
  finished ();
}

void
edit_modify::join (path pp) {
  path p= copy (pp);
  // cout << "Join at " << p << "\n";
  if (N(p)<1) fatal_error ("path too short in join", "editor::join");
  tree& st (subtree (et, path_up (p)));
  int  l1 = last_item (p);
  int  l2 = N (st[l1]);
  if (l2==0) l2= N (st[l1]->label);
  if (l1+1 >= N(st)) fatal_error ("invalid join", "editor::join");
  bool string_mode= (N(st[l1])==0) && (N(st[l1+1])==0);
  int len= string_mode? N(st[l1]->label): N(st[l1]);
  notify_undo ("split", path_down (p, len), "");

  int i;
  for (i=0; i<N(buf->vws); i++)
    ((tm_view) (buf->vws[i]))->ed->notify_join (et, p);

  if (string_mode) st[l1]->label << st[l1+1]->label;
  else st[l1] << st[l1+1]->a;
  st= ::remove (st, l1+1, 1);
  finished ();
}

void
edit_modify::ins_unary (path pp, string op) {
  path p= copy (pp);
  // cout << "Insert unary " << op << " at " << p << "\n";
  notify_undo ("rem_unary", p, "");

  int i;
  for (i=0; i<N(buf->vws); i++)
    ((tm_view) (buf->vws[i]))->ed->notify_ins_unary (et, p, op);

  tree& st (subtree (et, p));
  st= tree (op, st);
  finished ();
}

void
edit_modify::rem_unary (path pp) {
  path p= copy (pp);
  // cout << "Remove unary at " << p << "\n";
  tree& st (subtree (et, p));
  if (N(st)!=1) fatal_error ("not a unary tree", "editor::rem_unary");
  notify_undo ("ins_unary", p, st->label);

  int i;
  for (i=0; i<N(buf->vws); i++)
    ((tm_view) (buf->vws[i]))->ed->notify_rem_unary (et, p);

  st= st[0];
  finished ();
}

void
edit_modify::finished () {
  int i;
  for (i=0; i<N(buf->vws); i++)
    ((tm_view) (buf->vws[i]))->ed->post_notify (et);
}

/******************************************************************************
* Cursor handling after notification of changes in document
******************************************************************************/

void
edit_modify::notify_assign (tree& t, path p, tree u) { (void) u;
  if ((&t)!=(&et)) return;
  if (p<tp) tp= path_down (p, 0);

  if (nil (p)) typeset_notify_assign (N(u));
  else typeset_notify_modify (p->item);
}

void
edit_modify::notify_insert (tree& t, path p, tree u) {
  if ((&t)!=(&et)) return;
  if ((N(p)>=2) && path_inf (path_up (p), tp));
  else if (path_inf (p, tp) || (p <= tp)) {
    int nr= (N(u)==0? N(u->label): N(u));
    tp[N(p)-1] += nr;
  }
  
  if (atom (p)) typeset_notify_insert (p->item, N(u));
  else typeset_notify_modify (p->item);
}

void
edit_modify::notify_remove (tree& t, path p, int nr) {
  if ((&t)!=(&et)) return;
  if ((N(p)>=2) && path_inf (path_up (p), tp));
  else {
    tree& st (subtree (et, path_up (p)));
    path p2= path_add (p, nr, N(p)-1);
    if ((path_up (p) <= tp) && (N (st) == nr))
      tp= path_down (path_up (p), 0);
    else if (path_inf (p2, tp) || (p2 <= tp)) tp[N(p)-1] -= nr;
    else if (path_inf (p, tp)  || (p <= tp)) {
      if ((N(st)==0) || (last_item(p)==0)) tp= copy (p);
      else tp= path_down (path_dec (p), right_index (st [last_item(p)-1]));
    }
  }

  if (atom (p)) typeset_notify_remove (p->item, nr);
  else typeset_notify_modify (p->item);
}

void
edit_modify::notify_split (tree& t, path p) {
  if ((&t)!=(&et)) return;
  if (!(path_up (p, 2) <= path_up (tp)));
  else if (path_up (p, 2) == path_up (tp));
  else if (tp[N(p)-2]<p[N(p)-2]);
  else if (tp[N(p)-2]>p[N(p)-2]) tp[N(p)-2] ++;
  else if ((path_up (p) == path_up (tp)) &&
	   (N (subtree (et, path_up (p))) != 0)) {
    if (last_item (tp)==1) tp[N(p)-2] ++;
  }
  else if (tp[N(p)-1]<p[N(p)-1]);
  else {
    tp[N(p)-2] ++;
    tp[N(p)-1] -= last_item (p);
  }

  typeset_notify_modify (p->item);
  if (N(p) == 2) typeset_notify_insert (p->item, 1);
}

void
edit_modify::notify_join (tree& t, path p) {
  if ((&t)!=(&et)) return;
  tree& st (subtree (et, p));
  bool flag = (N(st)==0);
  int  extra= (flag? N(st->label): N(st));
  if (!(path_up (p) <= path_up (tp)));
  else if (path_up (p) == path_up (tp));
  else if (tp[N(p)-1] <= p[N(p)-1]);
  else {
    if ((tp[N(p)-1] == p[N(p)-1]+1) &&
	(flag || (path_inc (p) != path_up (tp)))) tp[N(p)] += extra;
    tp[N(p)-1] --;
  }

  typeset_notify_modify (p->item);
  if (N(p) == 1) typeset_notify_remove (p->item+1, 1);
}

void
edit_modify::notify_ins_unary (tree& t, path p, string op) { (void) op;
  if ((&t)!=(&et)) return;
  if (p <= path_up (tp)) {
    path add= path (0, tail (tp, N(p)));
    tp= merge (copy (p), add);
  }

  if (nil (p)) typeset_notify_assign (1);
  else typeset_notify_modify (p->item);
}

void
edit_modify::notify_rem_unary (tree& t, path p) {
  if ((&t)!=(&et)) return;
  if (p == path_up (tp)) {
    if (last_item (tp)==1)
      tp[N(tp)-1]= right_index (subtree (et, path_down (p, 0)));
  }
  else if (p <= path_up (tp)) {
    path add= tail (tp, N(p)+1);
    tp= merge (p, add);
  }

  if (nil (p)) typeset_notify_assign (N(et[0]));
  else typeset_notify_modify (p->item);
}

void
edit_modify::post_notify (tree& t) {
  if ((&t)!=(&et)) return;
  selection_clear ();
  notify_change (THE_TREE);
  set_cursor (correct_cursor (et, tp));
  /*
  cout << "et= " << et << "\n";
  cout << "tp= " << tp << "\n\n";
  */
}

/******************************************************************************
* undo and redo handling
******************************************************************************/

static tree
encode (string op, path p, tree t) {
  string s= copy (op);
  while (!nil (p)) {
    s << ";" << as_string (p->item);
    p= p->next;
  }
  if (t == "") return s;
  else return tree (TUPLE, s, t);
}

static void
decode (tree x, string& op, path& p, tree& t) {
  string s;
  if (N(x)==0) {
    s= x->label;
    t= "";
  }
  else {
    s= x[0]->label;
    t= x[1];
  }

  p= path ();
  int i=N(s);
  while (i>0) {
    int end= i;
    for (; i>0; i--)
      if (s[i-1]==';') break;
    if (i==0) op= s (i, end);
    else {
      p= path (as_int (s (i, end)), p);
      i--;
    }
  }
}

void
edit_modify::notify_undo (string op, path p, tree t) {
  // cout << "Undone by " << op << " " << t << " at " << p << "\n";
  tree x= encode (op, p, t);
  if (undo_flag) buf->redo= tree (BACKUP, x, buf->redo);
  else {
    if (!redo_flag) {
      if (buf->redo != "nil") {
	buf->redo_to_undo ();
	buf->mark_undo_block ();
      }
      if ((max_undo_depth > 0) && (buf->undo_depth >= (2*max_undo_depth)))
	buf->truncate_undos (max_undo_depth);
    }
    if ((op == "remove") &&
	(buf->needs_to_be_autosaved ()) &&
	(buf->undo != "nil") && (buf->undo[0] == "") &&
	(buf->undo[1] != "nil") &&
	(buf->undo[1][1] != "nil") && (buf->undo[1][1][0] == ""))
      {
	string op2; path p2; tree t2;
	decode (buf->undo[1][0], op2, p2, t2);
	if ((op2 == "remove") && (p == path_add (p2, as_int (t2->label)))) {
	  buf->unmark_undo_block ();
	  buf->undo= buf->undo [1];
	  int nr= as_int (t2->label)+ as_int (t->label);
	  x= encode (op, p2, as_string (nr));
	}
      }
    buf->undo= tree (BACKUP, x, buf->undo);
  }
  /*
  cout << "undo tree : " << buf->undo << "\n";
  cout << "redo tree : " << buf->redo << "\n";
  cout << "exdo tree : " << buf->exdo << "\n";
  cout << "undo depth: " << buf->undo_depth << "\n";
  cout << "redo depth: " << buf->redo_depth << "\n";
  cout << "last save : " << buf->last_save << "\n";
  */
}

void
edit_modify::undo () {
  buf->unmark_undo_block ();
  if (buf->undo == "nil") {
    set_message ("No more undo information available", "undo");
    return;
  }
  buf->mark_redo_block ();
  while ((buf->undo != "nil") && (buf->undo[0] != "")) {
    tree x= buf->undo[0];
    buf->undo= buf->undo[1];
    undo_flag= TRUE;
    buf->exdo= tree (BACKUP, copy (x), buf->exdo);
    perform_undo_redo (x);
    undo_flag= FALSE;
  }
  buf->unmark_undo_block ();
  if (buf->undo_depth == buf->last_save) {
    cerr << '\a';
    set_message ("Your document is back in its original state", "undo");
  }
}

void
edit_modify::redo () {
  buf->unmark_redo_block ();
  if (buf->redo == "nil") {
    set_message ("No more redo information available", "redo");
    return;
  }
  buf->mark_undo_block ();
  while ((buf->redo != "nil") && (buf->redo[0] != "")) {
    tree x= buf->redo[0];
    buf->redo= buf->redo[1];
    buf->exdo= buf->exdo[1];
    redo_flag= TRUE;
    perform_undo_redo (x);
    redo_flag= FALSE;
  }
  buf->unmark_redo_block ();
  if (buf->undo_depth == buf->last_save) {
    cerr << '\a';
    set_message ("Your document is back in its original state", "redo");
  }
}

void
edit_modify::perform_undo_redo (tree x) {
  string op; path p; tree t;
  decode (x, op, p, t);

  if (op == "assign") {
    assign (p, t);
    set_cursor (end (et, p));
  }
  else if (op == "insert") {
    insert (p, t);
    if (N(t)==0) set_cursor (path_add (p, N(t->label)));
    else set_cursor (end (et, path_add (p, N(t)-1)));
  }
  else if (op == "remove") {
    if (N (subtree (et, path_up (p))) == 0) {
      remove (p, as_int (t->label));
      set_cursor (p);
    }
    else {
      remove (p, as_int (t->label));
      if (last_item (p) == 0) set_cursor (start (et, p));
      else set_cursor (end (et, path_dec (p)));
    }
  }
  else if (op == "split") {
    split (p);
    path q= path_inc (path_up (p));
    if (N(subtree (et, q))==0) set_cursor (path_down (q, 0));
    else set_cursor (start (et, path_down (q, 0)));
  }
  else if (op == "join") {
    tree& st1 (subtree (et, p));
    tree& st2 (subtree (et, path_inc (p)));
    if ((N(st1)==0) && (N(st2)==0)) {
      int last= N(st1->label);
      join (p);
      set_cursor (path_down (p, last));
    }
    else {
      int  last= N(st1)-1;
      join (p);
      if (last == -1) set_cursor (start (et, p));
      else set_cursor (end (et, path_down (p, last)));
    }
  }
  else if (op == "ins_unary") {
    if (p < tp) ins_unary (p, t->label);
    else {
      ins_unary (p, t->label);
      set_cursor (end (et, path_down (p, 0)));
    }
  }
  else if (op == "rem_unary") {
    if (path_down (p, 0) < tp) rem_unary (p);
    else if (tp == path_down (p, 0)) {
      rem_unary (p);
      set_cursor (start (et, p));
    }
    else {
      rem_unary (p);
      set_cursor (end (et, p));
    }
  }
}

#endmodule // code_notify
