
/******************************************************************************
* MODULE     : edit_interface.gen.h
* DESCRIPTION: the interface for TeXmacs
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <edit_cursor.gen.h>
#include <edit_typeset.gen.h>
#include <window.gen.h>
#include <Widget/basic_widget.gen.h>
#include <Event/attribute_event.gen.h>
#include <timer.gen.h>

#module edit_interface
#import edit_cursor
#import edit_typeset
#import window
#import basic_widget
#import attribute_event
#import timer
#import list (path)

#define INPUT_NORMAL      0
#define INPUT_SEARCH      1
#define INPUT_REPLACE     2

class server_rep;

class edit_interface:
  virtual public edit_cursor,
  virtual public edit_typeset,
  public basic_widget_rep
{

protected:
  int         env_change;    // which things have been changed ?
  time_t      last_change;   // time of last processed change
  time_t      last_update;   // time of last update of menu, icons and footer
  int         con_status;    // 0: inexistent, 1: waiting, 2: active
  string      con_name;      // name of connected application
  string      con_session;   // name of computer algebra session
  bool        got_focus;     // do we have keyboard focus ?
  string      sh_s;          // current string for shorthands
  int         sh_len;        // length of translation
  scheme_tree main_menu;     // the current main menu
  scheme_tree icon_menu[3];  // the icon menus
  scheme_tree popup_menu;    // the current popup menu
  window      popup_win;     // the current popup window
  string      message_l;     // a left message to display
  string      message_r;     // a right message to display
  int         sfactor;       // the current shrinking factor
  SI          pixel;         // sfactor*PIXEL

protected:
  int         input_mode;    // INPUT_NORMAL, INPUT_SEARCH, INPUT_REPLACE
  bool        forward;       // forward search or replace ?
  string      search_mode;   // only search matches in search_mode mode
  string      search_lan;    // only search matches in search_lan language
  path        search_at;     // current search position
  path        search_end;    // position of the end of a match
  tree        search_what;   // search tree
  list<path>  search_stack;  // last search positions
  tree        keybd_stack;   // last key presses
  tree        replace_by;    // replace tree
  int         nr_replaced;   // number of replaced occurrences

public:
  server_rep*    sv;     // the underlying texmacs server

  edit_interface (server_rep* sv, display dis, tm_buffer buf);
  ~edit_interface ();
  operator tree ();
  void suspend ();
  void resume ();

  /* extern connections */
  void update_connection ();
  void connect ();
  void process_extern_input ();
  void feed_input (tree t);
  void lazy_shell_command (string s);
  bool busy_connection ();
  void interrupt_connection ();
  void stop_connection ();

  /* routines for dealing with shrinked coordinates */
  void set_shrinking_factor (int sf);
  void invalidate (SI x1, SI y1, SI x2, SI y2);
  void get_visible (SI& x1, SI& y1, SI& x2, SI& y2);
  SI   get_window_height ();
  void scroll_to (SI x, SI y1);
  void set_extents (SI x1, SI y1, SI x2, SI y2);

  /* repainting the window */
  void draw_text (repaint_event ev);
  void draw_context (repaint_event ev);
  void draw_cursor (ps_device dev);
  void draw_selection (ps_device dev);

  /* handle changes */
  void notify_change (int changed);
  bool has_changed (int question);
  void process_style (tree style);
  void cursor_visible ();
  void selection_visible ();
  void apply_changes ();

  /* miscellaneous */
  bool get_latex_command (string which, string& help, scheme_tree& p);
  void set_main_menu (scheme_tree p);
  void set_popup_menu (scheme_tree p);
  void set_icon_menu (int which, scheme_tree p);

  /* keyboard handling */
  void set_input_normal ();
  void key_press (string key);
  void emulate_keyboard (string keys, string action= "");
  void show_keymaps ();

  /* mouse handling */
  void mouse_press (string s, SI x, SI y);
  void mouse_select (SI x, SI y);
  void mouse_paste (SI x, SI y);
  void mouse_drag (SI x, SI y);
  void mouse_adjust (SI x, SI y);
  void selection_set_start ();
  void selection_set_end ();
  bool selection_active ();
  void selection_get (path& start, path& end);
  void selection_set (tree t);
  void selection_clear ();
  void selection_copy ();
  void selection_paste ();

  /* the footer */
  string compute_text_footer (tree st);
  string compute_operation_footer (tree st);
  string compute_compound_footer (tree t, path p);
  bool   set_latex_footer (tree st);
  void   set_left_footer (string l);
  void   append_left_footer (string& s, string env_var);
  void   set_left_footer ();
  void   set_right_footer (string r);
  void   set_right_footer ();
  void   set_footer ();
  void   set_message (string l, string r= "");
  void   interactive (scheme_tree args, scheme_tree cmd);

  /* search and replace */
  path   test (path p, tree t);
  void   step_ascend (bool forward);
  void   step_descend (bool forward);
  void   step_horizontal (bool forward);
  void   next_match (bool forward);
  void   search_start (bool forward= TRUE);
  void   search_next (bool forward);
  void   search_button_next ();
  bool   search_keypress (string s);
  void   replace_start (tree what, tree by, bool forward= TRUE);
  void   replace_next ();
  bool   replace_keypress (string s);

  /* event handlers */
  void handle_get_size (get_size_event ev);
  void handle_attach_window (attach_window_event ev);
  void handle_repaint (repaint_event ev);
  void handle_keypress (keypress_event ev);
  void handle_mouse (mouse_event ev);
  void handle_keyboard_focus (keyboard_focus_event ev);
  void handle_resize (resize_event ev);

  friend class interactive_command_rep;
  friend class tm_window_rep;
  friend class tm_project_rep;
};

#define SERVER(cmd) {                    \
  tm_view temp_vw= sv->get_view (FALSE); \
  focus_on_this_editor ();               \
  sv->cmd;                               \
  sv->set_view (temp_vw);                \
}

#endmodule // edit_interface
